import java.io.*;
import java.util.*;

import org.jdom2.*;



/**
 * Takes a featuremodel.xml file generated using FeatureModelConfigurator and a given .gazebo.xacro and urdf.xacro file.
 * The .gazebo.xacro and .urdf.xacro files are modified to support the features specified in the featuremodel.xml file.
 */
public class Adapter extends XMLHelper {
	
	/**
	 * Obtains an attribute value of the input/ output entries specified in adapter.config.
	 * 
	 * @param doc		The JDOM Document (adapter.config)
	 * @param entry		String formed as "{input or output}/{entity}"
	 * @param val		Attribute name as string
	 * @return path as specified in adapter.config
	 */
	private static String getConfigVal(Document doc, String str, String val) {
		List<Attribute> valList = xQueryAttributes(doc, "//"+str+"/@"+val);
		if(valList.size()!=1) {
			System.err.println("Error while reading input/ output "+val+" value from adapter.config.");
			System.exit(-1);
		}

		return valList.get(0).getValue();
	}
	
	/**
	 * Call this to run Adapter.
	 */
	//public static void main(String[] args) {
	public Adapter() {
		System.out.println("Adapter started.");
		
		try {
			final String SC_ENDING = ".xml";
			final String GX_ENDING = ".gazebo.xacro";
			final String UX_ENDING = ".urdf.xacro";
			
			//######### Input ###############################			
			Document adapterConfig = readXml("./adapter.config");
			
			//Get adapter.config values.
			String in_fmPath = getConfigVal(adapterConfig, "input/featuremodel", "path");
			String in_scPath = getConfigVal(adapterConfig, "input/sensorconfig", "path");
			String in_gxPath = getConfigVal(adapterConfig, "input/gazebo_xacro", "path");
			String in_uxPath = getConfigVal(adapterConfig, "input/urdf_xacro", "path");
			String out_gxPath = getConfigVal(adapterConfig, "output/gazebo_xacro", "path");
			String out_uxPath = getConfigVal(adapterConfig, "output/urdf_xacro", "path");
			String out_bnPath = getConfigVal(adapterConfig, "output/botname", "path");

			String in_fmName = getConfigVal(adapterConfig, "input/featuremodel", "name");
			String in_gxName = "";
			String in_uxName = "";
			String out_gxName = getConfigVal(adapterConfig, "output/gazebo_xacro", "name");
			String out_uxName = getConfigVal(adapterConfig, "output/urdf_xacro", "name");
			String out_bnName = getConfigVal(adapterConfig, "output/botname", "name");
			
			String unwantedPrefix = getConfigVal(adapterConfig, "output/botname", "remove_prefix");
			String unwantedSuffix = getConfigVal(adapterConfig, "output/botname", "remove_suffix");
			
			//Declare Documents.
			Document featureModel = readXml(in_fmPath+in_fmName);
			List<Document> sensorConfigList = new ArrayList<Document>();
			Document gazeboModel = null;
			Document urdfModel = null;
			
			//Get the selected mode (either NO_CHANGE or Custom) as String.
			String mode = "";
			
			List<Attribute> modeList = xQueryAttributes(featureModel, "//mode/@param");
			if(modeList.size()==1) {
				mode = modeList.get(0).getValue();
			}
			else {
				System.err.println("Invalid input in \'"+in_fmName+"\': NO_CHANGE/ Custom parameter could not be parsed.");
				System.exit(-1);
			}
			
			//Get all selected sensorconfig files as List<Document>.
			switch(mode) {
				case "NO_CHANGE":
					break;
				case "Custom":
					List<Attribute> sensorList = xQueryAttributes(featureModel, "//sensor/@name");
					
					List<String> scList = getFileNames(in_scPath);
					for(Attribute sensor : sensorList) {
						boolean inDirectory = false;
						for(String fileName : scList) {
							if(fileName.equals(sensor.getValue()+SC_ENDING)) inDirectory = true;
						}
						if(!inDirectory) {
							System.err.println("Cannot handle \'"+sensor.getValue()+"\': No input file known.");
							System.exit(-1);
						}
					}
					
					for(Attribute sensor : sensorList) {
						sensorConfigList.add(readXml(in_scPath+sensor.getValue()+SC_ENDING)); 
					}
					break;
				default:
					System.err.println("Invalid input in \'"+in_fmName+"\': NO_CHANGE/ Custom parameter could not be parsed.");
					System.exit(-1);
			}
			
			//Get the name of the selected bot as String.
			String botName = "";
			
			List<Attribute> botList = xQueryAttributes(featureModel, "//templates/@name");
			if(botList.size()==1) {
				botName = botList.get(0).getValue();
			}
			else {
				System.err.println("Invalid input in \'"+in_fmName+"\': No or multiple bots selected.");
				System.exit(-1);
			}

			//Get the selected .gazebo.xacro file as Document.
			in_gxName = botName+GX_ENDING;
			
			List<String> gxList = getFileNames(in_gxPath);
			boolean inDirectory = false;
			for(String fileName : gxList) {
				if(fileName.equals(in_gxName)) inDirectory = true;
			}
			if(!inDirectory) {
				System.err.println("Cannot handle \'"+botName+"\': No \'"+GX_ENDING+"\' input file known.");
				System.exit(-1);
			}
			
			gazeboModel = readXml(in_gxPath+in_gxName);
			
			//Get the corresponding .urdf.xacro file as Document.
			in_uxName = botName+UX_ENDING;
			
			List<String> urdfList = getFileNames(in_uxPath);
			inDirectory = false;
			for(String fileName : urdfList) {
				if(fileName.equals(in_uxName)) inDirectory = true;
			}
			if(!inDirectory) {
				System.err.println("Cannot handle \'"+botName+"\': No \'"+UX_ENDING+"\' input file known.");
				System.exit(-1);
			}
			
			urdfModel = readXml(in_uxPath+in_uxName);
			//###############################################
			
			
			//######### Operate on .gazebo.xacro ############
			switch(mode) {
				case "NO_CHANGE":
					break;
				case "Custom":
					//Remove all existing sensors in template.
					List<Element> removableSensorsList = xQueryElements(gazeboModel, "//sensor");
					
					for(Element sensor : removableSensorsList) {	
						Element entry = sensor.getParentElement();
						if(entry.getName().equals("gazebo")) {
							entry.detach();
						}
						else {
							System.out.println("Found a <sensor> element in \'"+in_gxName+
								"\' that does not have a <gazebo> element as immediate parent. It was not removed, as this input is unexpected.");	
						}
					}
					
					//Insert selected sensors into template.
					for(Document sensorConfig : sensorConfigList) {	
						List<Element> sensorList = xQueryElements(sensorConfig, "//gazebo");
						if(sensorList.size()==1) {
							Element sensor = sensorList.get(0);
							
							//Warn if a sensor still contains $(arg ...) values.
							List<Element> tagList = xQueryElements(sensor, "//*");
							for(Element tag : tagList) {	
								String text = tag.getText();
								if(text.contains("$")) {
									System.out.println("Warning: a sensorconfig file still contains an abstract value: \'"+text+
										"\'. It should be replaced with a set value, or the output .gazebo.xacro will likely be broken.");	
								}
							}
							
							sensor.detach();
							gazeboModel.getRootElement().addContent(sensor);
						}
						else {
							System.err.println("Invalid input. At least one sensorconfig file contains no or multiple <gazebo> entries.");
							System.exit(-1);
						}
					}
					break;
				default:
					System.err.println("Invalid input in \'"+in_fmName+"\': NO_CHANGE/ Custom parameter could not be parsed.");
					System.exit(-1);
			}
			//###############################################
			
			
			//######### Operate on .urdf.xacro ##############
			final Namespace NS = urdfModel.getRootElement().getNamespace("xacro");
			
			//Set <xacro:include> tags
			List<Element> includeList = xQueryElements(urdfModel, "//xacro:include", NS);
			if(includeList.size()!=2) {
				System.err.println("Invalid input in \'"+in_uxName+"\': Expected exactly 2 xacro:include tags.");
				System.exit(-1);
			}
			String gxDestDirectoryName = (new File(out_gxPath)).getName();
			//includeList.get(0).setAttribute("filename", "$(find "+CommonPropDestDirectoryName+")/common_properties.xacro");//TODO
			includeList.get(1).setAttribute("filename", "$(find "+gxDestDirectoryName+")/"+out_gxName);
			
			switch(mode) {
				case "NO_CHANGE":
					break;
				case "Custom":
					//TODO
					//If the urdf.xacro would be modified (to have an accurate visual model with the sensors attached to it), 
					//this needs to happen here. (Delete or Insert <link>/ <joint> tags.)
					//This would require separate files with <link>/ <joint> tags for each bot for the Adapter to access...
					break;
				default:
					System.err.println("Invalid input in \'"+in_fmName+"\': NO_CHANGE/ Custom parameter could not be parsed.");
					System.exit(-1);
			}
			//###############################################
			
			
			//######### Output result #######################
			//Output model files.
			saveAsXml(gazeboModel, out_gxPath+out_gxName);
			
			saveAsXml(urdfModel, out_uxPath+out_uxName);
			
			//Output botname.
			String alteredBotName = botName;
			if(alteredBotName.startsWith(unwantedPrefix)) {
				alteredBotName = alteredBotName.substring(unwantedPrefix.length(), alteredBotName.length());
			}
			else {
				System.out.println("Warning: Prefix  \'"+unwantedPrefix+" could not be removed from bot name.");
			}
			if(alteredBotName.endsWith(unwantedSuffix)) {
				alteredBotName = alteredBotName.substring(0, alteredBotName.length() - unwantedSuffix.length());
			}
			else {
				System.out.println("Warning: Prefix  \'"+unwantedSuffix+" could not be removed from bot name.");
			}

			BufferedWriter bufferedWriter = new BufferedWriter(new FileWriter(out_bnPath+out_bnName));
			bufferedWriter.write(alteredBotName);
			bufferedWriter.close();
			//###############################################
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		System.out.println("Adapter finished.");
	}
}



